#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2018 Cosmos Coin Developers, https://cosmoscoin.co/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

from . import vps
from . import core

import argparse
import configparser

CONFIG_FILENAME = os.path.join(os.path.dirname(__file__), "conf", "config.ini")

DESCRIPTION = "End to end script to setup a masternode"

ROOT_USER = "root"

def getConfig():
    """Parse configuration file.
    
    Returns:
        dict: Dictionary containing the options parsed.
    """
    config = configparser.ConfigParser()
    config.read(CONFIG_FILENAME)
    
    return config

def begin(args):
    """Wrapper function that starts the application with the given arguments.
    
    Args:
        args (obj): Object containing the command line arguments parsed.
    """
    try:
        # Parse file configuration
        config = getConfig()

        # Ensure all local requirements met (e.g. wallet installed)
        core.checkPrerequisites(config)
    
        # Setup VPS - Update packages, install binaries
        vps.setup(args.vps, ROOT_USER, args.password, config)
    
        # Setup masternode locally
        core.setup(args.vps, ROOT_USER, args.password, args.name, config)
        
    except Exception as e:
        print("Masternode setup failed. Reason: {0}.".format(str(e)))  
        raise e
    
def setup():
    """Program entrypoint. This function will parse all program arguments and start the application.
    """
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    
    parser.add_argument("--name", action="store", required=True, help="The name to be given to the masternode")
    parser.add_argument("--vps", action="store", required=True, help="The IP address of the VPS server to be used")
    parser.add_argument("--password", action="store", required=True, help="The root password for the VPS provided")
    
    args = parser.parse_args()    
    begin(args)
    
if __name__ == "__main__":
    setup()


#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2018 Cosmos Coin Developers, https://cosmoscoin.co/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import glob
import time
import json

import pprint
import string
import random

import getpass
import subprocess

from . import vps
from . import daemon

from os import environ

RPC_USER_LENGTH = 32
RPC_PASSWORD_LENGTH = 32

CONF_TEMPLATE_FILE = os.path.join(os.path.dirname(__file__), "conf", "conf.template")
ACTIVATION_STRING = "waiting for remote activation"

def checkIfEnvironmentDefined(envHome, envUser):
    """Checks if the required environment variables are defined.
    
    Args:
        envHome (str): The name of the home environment variable.
        envUser (str): The name of the user environment variable.
    """
    if envHome not in environ:
        raise ValueError("Please set {0} environment variable".format(envHome))
        
    elif envUser not in environ:
        raise ValueError("Please set {0} environment variable".format(envUser))
       
def checkIfWalletInstalled(cli, daemonCli):
    """Check if the local wallet is installed.

    Args:
        cli (str): The full path of the local cli binary.
        daemonCli (str): The full path of the local daemon binary.
    """
    if not glob.glob("{0}*".format(cli)):
        raise ValueError("Unable to find file: {0}, please check your installation".format(cli))
    
    elif not glob.glob("{0}*".format(daemonCli)):
        raise ValueError("Unable to find file: {0}, please check your installation".format(daemonCli))

def setupWallet(walletConfFile):
    """This function will setup the local wallet requirements.
    
    Args:
        walletConfFile (str): Full path to wallet conf file.
    """
    print("Setup wallet configuration..")
    if "rpcuser" not in open(walletConfFile).read():
        print("Creating rpcuser setting..")
        with open(walletConfFile, "a") as f:
            f.write("rpcuser={0}\n".format(generateRandomString(RPC_USER_LENGTH)))
            
    if "rpcpassword" not in open(walletConfFile).read():
        print("Creating rpcpassword setting..")
        with open(walletConfFile, "a") as f:
            f.write("rpcpassword={0}\n".format(generateRandomString(RPC_PASSWORD_LENGTH)))
    
    print("")
    
def startLocalDaemon(daemonCli):
    """Start the local daemon. Required to use the coin cli.
    
    Args:
        daemonCli (str): The full path of the local daemon binary.
        
    Returns:
        Obj: The process associated with the daemon that was started.
    """    
    print("Starting local daemon..")
    
    try:
        proc = daemon.start(daemonCli)
        
        # Ensure that the daemon did not fail and is still running
        proc.communicate(timeout=20)
    except subprocess.TimeoutExpired:
        pass
    else:
        raise ValueError("Failed to start local daemon")
        
    print("")
    return proc

def stopLocalDaemon(cli):
    """Stop the local daemon.
    
    Args:
        cli (str): The full path of the local cli binary.
    """    
    print("Stopping local daemon..")
    try:
        daemon.stop(cli)
    except:
        raise ValueError("Failed to stop local daemon")
        
    print("")

def pollForWalletSync(cli):
    """Poll for the wallet to be fully synced (on daemon).
    
    Utilizes the blockchain info RPC call to check that verification in progress
    is at least 1.00. This field is an estimate of the synchronization status. 
    Checking this number ensures we are in sync with a good confidence level.
    
    Args:
        cli (str): The full path of the local cli binary.
    """    
    print("Wait for wallet to by synchronized..")
    blockchaininfo = daemon.getBlockchainInfo(cli)
    blockchainvalues = json.loads(blockchaininfo)

    # Loop continuously until we know we are synced
    while blockchainvalues["verificationprogress"] < 1:
        blockchaininfo = daemon.getBlockchainInfo(cli)
        blockchainvalues = json.loads(blockchaininfo)

        print("Progress.. {0}%".format(blockchainvalues["verificationprogress"] * 100))
        time.sleep(5)
     
    print("")
            
def getTotalUnlockedBalance(cli):
    """Get the total balance in wallet without accounting for locked coins (i.e. from masternodes)
    
    Args:
        cli (str): The full path of the local cli binary.
        
    Returns:
        Int: The total unlocked balance present.
    """    
    output = daemon.listUnspent(cli)
    values = json.loads(output)
    
    return sum(input['amount'] for input in values)
    
def unlockWallet(cli):
    """Unlocks the wallet temporarily so that other operations may be performed.
    
    Args:
        cli (str): The full path of the local cli binary.
    """    
    done = False
    
    # Loop until the correct password is seen
    while done == False:
        try:
            passphrase = getpass.getpass(prompt="Please enter you wallet passphrase: ")
            output = daemon.unlockWallet(cli, passphrase)
        except:
            pass
            print("Incorrect passphrase, please try again..")
        else:
            done = True
    
    print("")
    
def sendCollateralToAddress(cli, address, collateral):
    """Send the collateral amount to the specified address.
    
    Args:
        cli (str): The full path of the local cli binary.
        address (str): The address to be used for the masternode.
        collateral (int): The collateral amount to be sent.
        
    Returns:
        String: The txid associated with this transaction.
    """    
    try:
        # If the wallet is not encrypted, no unlocking is necessary
        txid = daemon.sendToAddress(cli, address, collateral)
    
    except:
        print("Wallet is locked, please unlock your wallet..")
            
        unlockWallet(cli)
        txid = daemon.sendToAddress(cli, address, collateral)
    
    return txid

def getMasternodeOutput(cli, txid):
    """Get the masternode output containing the masternode txid.
    
    Args:
        cli (str): The full path of the local cli binary.
        txid (str): The txid associated with the collateral.
        
    Returns:
        Dict: A dictionary containing the masternode output
    """         
    output = daemon.getMasternodeOutputs(cli)
    values = json.loads(output)
    
    txValues = [value for value in values if value['txhash'] == txid]
    
    # There should be a single match, check to be sure
    if len(txValues) != 1:
        raise ValueError("Transaction: {0} was not found in masternode outputs".format(txid))
    
    return txValues[0]
    
def setupMasternodeTransaction(cli, label, collateral):
    """Performs a setup of the transactions required to create a masternode.
    
    Specifically:
    1. Check balance
    2. Generate address
    3. Send collateral to address
    4. Get masternode output
    5. Get masternode key
    
    Args:
        cli (str): The full path of the local cli binary.
        label (str): The label to be used when generating a new address.
        collateral (int): The collateral amount to be sent.
        
    Returns:
        2-Tuple: Tuple containing the masternode output and masternode key
    """         
    print("Setup masternode transaction..\n")
    
    # First check we have enough balance
    balance = getTotalUnlockedBalance(cli)
    print("Collateral required: {0}, total unlocked balance: {1}".format(collateral, balance))
    
    if balance < collateral:
        raise ValueError("Insufficient funds")    
    
    # Generate new address for masternode
    address = daemon.generateNewAddress(cli, label)
    print("Send collateral to new address: {{{0}: {1}}}".format(label, address))
    
    # Send collateral to new address
    txid = sendCollateralToAddress(cli, address, collateral)
    print("Sent collateral successfully, txid: {0}\n".format(txid))
    
    # Get masternode output associated with collateral
    print("Get masternode output..")
    masternodeOutput = getMasternodeOutput(cli, txid)
    print("Masternode output:\n{0}\n".format(masternodeOutput))
    
    # Get masternode key to be used
    print("Get masternode key..")
    masternodeKey = daemon.generateMasternodeKey(cli)
    print("Masternode key: {0}\n".format(masternodeKey))
    
    return (masternodeOutput, masternodeKey)

def generateRandomString(length):
    """Generate a random string of a specified length. To be used in rpc information.
    
    Args:
        length (int): Length of string to be generated.
        
    Returns:
        String: String that was generated.
    """
    choice = string.ascii_uppercase + string.ascii_lowercase + string.digits
    return ''.join(random.SystemRandom().choice(choice) for _ in range(length))

def setupVpsConfFile(server, user, password, confFile, masternodeKey):
    """Setup the vps masternode configuration file.
    
    Args:
        server (str): The IP address of the server to connect to.
        user (str): The username to be used in the connection.
        password (str): The password associated with the user.
        confFile (str): The full path to the configuration file to be used.
        masternodeKey (str): The masternode key associated with this node.
    """
    print("Setup VPS conf file..")

    values = dict()
    
    # Generate random values for user and password
    values["rpcuser"] = generateRandomString(RPC_USER_LENGTH)
    values["rpcpassword"] = generateRandomString(RPC_PASSWORD_LENGTH)
    
    values["externalip"] = server
    values["masternodepivkey"] = masternodeKey  

    # Parse and replace template file with values
    source = ""
    with open(CONF_TEMPLATE_FILE) as template:
        source = string.Template(template.read())
    
    sourceFile = source.substitute(values)
    
    print("Generated configuration file:")
    pprint.pprint(sourceFile)
    print("")
    
    # Send configuration file to vps
    print("Send configuration file to VPS..")
    vps.createFileWithContents(server, user, password, confFile, sourceFile)
    print("Configuration file sent successfully!\n")

def updateVpsPermissions(server, user, password, coinName):
    """Updates the folder permissions for the coin user so that the daemon may run.
    
    Args:
        server (str): The IP address of the server to connect to.
        user (str): The username to be used in the connection.
        password (str): The password associated with the user.
        coinName (str): The name of the coin.
    """
    print("Updating folder permissions for {0} user.".format(coinName))
    
    command = "chmod -R 777 /home/{0}".format(coinName)
    vps.sendSingleCommand(server, user, password, command)
    
    print("")
    
def stopVpsDaemon(cli, daemonCli, server, user, password, coinName): 
    """Stops the vps daemon if it's running.
    
    Args:
        cli (str): The name of the coin cli to be used.
        daemonCli (str): The name of the coin daemon to be used.
        server (str): The IP address of the server to connect to.
        user (str): The username to be used in the connection.
        password (str): The password associated with the user.
        coinName (str): The name of the coin.
    """  
    print("Stopping daemon on VPS (if running)..")
    
    # Open SSH connection
    channel = vps.openChannel(server, user, password)    

    try: 
        if vps.isProcessRunning(channel, daemonCli):
            # Stop the process if it's currently running
            command = "su -c \"{0} stop\" {1}".format(cli, coinName)
            output = vps.sendCommand(channel, command)

            # Allow enough time for process to terminate
            time.sleep(20)
    finally:
        # Close ssh connection
        vps.closeChannel(channel)

    print("")
    
def startVpsDaemon(daemonCli, server, user, password, coinName):
    """Starts the vps daemon.
    
    Args:
        daemonCli (str): The name of the coin daemon to be used.
        server (str): The IP address of the server to connect to.
        user (str): The username to be used in the connection.
        password (str): The password associated with the user.
        coinName (str): The name of the coin.
    """  
    print("Starting daemon on VPS..")
    
    # Open SSH connection
    channel = vps.openChannel(server, user, password)

    try:
        command = "su -c \"{0} -daemon\" {1}".format(daemonCli, coinName)
        print(vps.sendCommand(channel, command))
        
        # Allow enough time for daemon to start
        time.sleep(20) 

        if not vps.isProcessRunning(channel, daemonCli):
            raise ValueError("Failed to start daemon on VPS")
        
    finally:
        # Close ssh connection
        vps.closeChannel(channel)

def clearVpsDebugFile(server, user, password, debugFile):
    """Removes the debug file specified.

    Args:
        server (str): The IP address of the server to connect to.
        user (str): The username to be used in the connection.
        password (str): The password associated with the user.
        debugFile (str): The name of the file to be removed.
    """
    print("Removing VPS debug file: {0}".format(debugFile))
    
    command = "rm -rf {0}".format(debugFile)
    vps.sendSingleCommand(server, user, password, command)
    
    print("")
    
def pollForVpsDaemonActivationReady(server, user, password, debugFile):
    """Waits until the daemon is ready to be activated.

    Args:
        server (str): The IP address of the server to connect to.
        user (str): The username to be used in the connection.
        password (str): The password associated with the user.
        debugFile (str): The name of the file to search.
    """
    done = False
    command = "grep \"{0}\" {1} > /dev/null".format(ACTIVATION_STRING, debugFile)
        
    while done == False:
        try:
            print("Wait until VPS daemon is activation ready.. This may take some time..")
            vps.sendSingleCommand(server, user, password, command)
        except:
            time.sleep(10)
        else:
            done = True
    
    print("")
    
def setupVpsMasternode(cli, daemonCli, server, user, password, confFile, masternodeKey, coinName, debugFile):
    """Sets up the masternode on the VPS.  Specifically:
    
    1. Stop daemon (if necessary).
    2. Copy masternode conf file to vps.
    3. Start daemon.
    
    Args:
        cli (str): The name of the coin cli to be used.
        daemonCli (str): The name of the coin daemon to be used.
        server (str): The IP address of the server to connect to.
        user (str): The username to be used in the connection.
        password (str): The password associated with the user.
        confFile (str): The full path to the configuration file to be used.
        masternodeKey (str): The masternode key associated with this node. 
        coinName (str): The name of the coin.
    """  
    # We must stop the current daemon to prevent issues with the conf file
    stopVpsDaemon(cli, daemonCli, server, user, password, coinName)
        
    # Load new conf file
    setupVpsConfFile(server, user, password, confFile, masternodeKey)
    
    # We must update folder permissions so that daemon can run
    updateVpsPermissions(server, user, password, coinName)
    
    # Clear debug file so that it can be searched later
    clearVpsDebugFile(server, user, password, debugFile)
    
    # Start daemon
    startVpsDaemon(daemonCli, server, user, password, coinName)
    
    # Poll for daemon to be ready for activation
    # Note: It seems like the activation doesn't work outside of wallet so disabling for now..
    #pollForVpsDaemonActivationReady(server, user, password, debugFile)

def setupMasternodeConfFile(server, label, masternodeConfFile, masternodePort, masternodeOutput, masternodeKey):
    """Sets up the local masternode conf file with the masternode values.
        
    Args:
        server (str): The IP address of the server to connect to.
        label (str): The label to be used when generating a new address.
        masternodeConfFile (str): The full path to the masternode conf file.
        masternodePort (int): The port number associated with the masternode.
        masternodeOutput (dict): The masternode output containing the collateral txinfo.
        masternodeKey (str): The masternode key associated with this node.        
    """  
    print("Setup masternode conf file..")
    
    # Append to the existing file
    with open(masternodeConfFile, 'a') as f:
        line = "\n{0} {1}:{2} {3} {4} {5}".format(label, server, masternodePort, masternodeKey, masternodeOutput["txhash"], masternodeOutput["outputidx"])
        f.write(line)

def startMasternodeAlias(cli, label):
    """Start the masternode by its alias.
    
    Args:
        cli (str): The name of the coin cli to be used.
        label (str): The label to be used when generating a new address.
    """
    print("Start masternode..")
    
    try:
        # If the wallet is not encrypted, no unlocking is necessary
        daemon.masternodeStartAlias(cli, label)
    
    except:
        print("Wallet is locked, please unlock your wallet..")
            
        unlockWallet(cli)
        daemon.masternodeStartAlias(cli, label)    

    print("Masternode started successfully!")

def setupWalletForMasternode(cli, server, label, masternodeConfFile, masternodePort, masternodeOutput, masternodeKey):
    """Sets up the wallet for the newly created masternode.
        
    Args:
        cli (str): The name of the coin cli to be used.    
        server (str): The IP address of the server to connect to.
        label (str): The label to be used when generating a new address.
        masternodeConfFile (str): The full path to the masternode conf file.
        masternodePort (int): The port number associated with the masternode.        
        masternodeOutput (dict): The masternode output containing the collateral txinfo.
        masternodeKey (str): The masternode key associated with this node.        
    """  
    setupMasternodeConfFile(server, label, masternodeConfFile, masternodePort, masternodeOutput, masternodeKey)
    startMasternodeAlias(cli, label)

def getCoinBinaries(envHome, cliName, daemonName):
    """Gets the full paths of the binaries associated with this coin.
    
    Args:
        envHome (str): The name of the home environment variable.
        cliName (str): The name of the cli binary associated with this coin.
        daemonName (str): The name of the daemon binary associated with this coin.
        
    Returns:
        2-Tuple: Tuple containing the full paths as described.
    """
    homePath = environ.get(envHome)

    cli = os.path.join(homePath, "daemon", cliName)
    daemonCli = os.path.join(homePath, "daemon", daemonName)
    
    return (cli, daemonCli)

def getCoinFiles(envUser, walletConf, masternodeConf):    
    """Gets the full paths of the conf files associated with this coin.
    
    Args:
        envUser (str): The name of the user environment variable.
        walletConf (str): The name of the wallet conf file.
        masternodeConf (str): The name of the masternode conf file.
        
    Returns:
        2-Tuple: Tuple full paths as described.
    """
    userPath = environ.get(envUser)
    
    walletConfFile = os.path.join(userPath, walletConf)
    masternodeConfFile = os.path.join(userPath, masternodeConf)
    
    return (walletConfFile, masternodeConfFile)

def checkPrerequisites(config):
    """Checks that certain prerequisits are met before continuing. Specifically:
    
    1. Check that necessary environment variables are defined.
    2. Check that the local wallet is installed.
    
    Args:
        envUser (str): The name of the user environment variable.
        masternodeConf (str): The name of the masternode conf file.
    """
    print("Checking wallet requirements..")    
    
    cli, daemonCli = getCoinBinaries(config["Environment"]["Home"], config["Coin"]["Cli"], config["Coin"]["Daemon"])    
    checkIfEnvironmentDefined(config["Environment"]["Home"], config["Environment"]["User"])
    checkIfWalletInstalled(cli, daemonCli)    
    
    print("")
        
def setup(server, user, password, label, config):
    """Top level function associated with this module. Responsible for the core configuration
    of this masternode. Specifically it will:
    
    1. Start local daemon.
    2. Poll for wallet to be synced.
    3. Perform masternode transactions.
    4. Setup masternode config on vps.
    5. Setup the local wallet for the masternode.
    6. Stop local daemon.
    """
    # Determine binary and file names based on config
    cli, daemonCli = getCoinBinaries(config["Environment"]["Home"], config["Coin"]["Cli"], config["Coin"]["Daemon"])
    walletConfFile, masternodeConfFile = getCoinFiles(config["Environment"]["User"], config["Wallet"]["WalletConf"], config["Wallet"]["MasternodeConf"])

    setupWallet(walletConfFile)
    localDaemon = startLocalDaemon(daemonCli)   
    
    try:
        pollForWalletSync(cli)
        masternodeOutput, masternodeKey = setupMasternodeTransaction(cli, label, float(config["Coin"]["Collateral"]))
        
        vpsConfFile = "/home/{0}/{1}/{2}".format(config["Coin"]["Name"], config["VPS"]["DataDir"], config["VPS"]["ConfFile"])
        vpsDebugFile = "/home/{0}/{1}/{2}".format(config["Coin"]["Name"], config["VPS"]["DataDir"], config["VPS"]["DebugFile"])
        
        setupVpsMasternode(config["Coin"]["Cli"], config["Coin"]["Daemon"], server, user, password, vpsConfFile, masternodeKey, config["Coin"]["Name"], vpsDebugFile)
        
        setupWalletForMasternode(cli, server, label, masternodeConfFile, int(config["Coin"]["Port"]), masternodeOutput, masternodeKey)

    finally:
        stopLocalDaemon(cli)
        localDaemon.wait()
    


#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2018 Cosmos Coin Developers, https://cosmoscoin.co/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import subprocess

DEFAULT_DECODE = "utf-8"
WALLET_LOCK_TIMEOUT_SEC = 60

DAEMON_STOP_COMMAND = "{0} stop"
DAEMON_START_COMMAND = "{0} -daemon"

CLI_GET_BALANCE = "{0} getbalance"
CLI_LIST_UNSPENT = "{0} listunspent"

CLI_GET_BLOCKCHAIN_INFO = "{0} getblockchaininfo"
CLI_UNLOCK_WALLET = "{0} walletpassphrase {1} {2}"

CLI_SEND_TO_ADDRESS = "{0} sendtoaddress {1} {2}"
CLI_GENERATE_NEW_ADDRESS = "{0} getnewaddress {1}"

CLI_MASTERNODE_GENKEY = "{0} masternode genkey"
CLI_MASTERNODE_OUTPUTS = "{0} masternode outputs"
  
CLI_MASTERNODE_START_ALIAS = "{0} masternode start-alias {1}"  

def start(daemon):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        daemon (str): Full path to daemon binary associated with coin.
        
    Returns:
        String: String containing the command output.
    """
    command = DAEMON_START_COMMAND.format(daemon).split(" ")
    return subprocess.Popen(command, stdout=subprocess.PIPE)
    
def stop(cli):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        
    Returns:
        String: String containing the command output.        
    """
    command = DAEMON_STOP_COMMAND.format(cli)
    return subprocess.check_output(command).decode(DEFAULT_DECODE)

def getBlockchainInfo(cli):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_GET_BLOCKCHAIN_INFO.format(cli)
    return subprocess.check_output(command).decode(DEFAULT_DECODE)
    
def generateNewAddress(cli, label):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        label (str): Label associated with the address to be generated.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_GENERATE_NEW_ADDRESS.format(cli, label)
    return subprocess.check_output(command).decode(DEFAULT_DECODE).strip()
    
def unlockWallet(cli, passphrase):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        passphrase (str): Passphrase to be used in wallet unlock.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_UNLOCK_WALLET.format(cli, passphrase, WALLET_LOCK_TIMEOUT_SEC)
    return subprocess.check_output(command, stderr=subprocess.STDOUT).decode(DEFAULT_DECODE)
            
def sendToAddress(cli, address, amount):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        address (str): Address to be used for sending.
        amount (int): Amount of coins to be sent.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_SEND_TO_ADDRESS.format(cli, address, amount)
    return subprocess.check_output(command, stderr=subprocess.STDOUT).decode(DEFAULT_DECODE).strip()

def getTotalBalance(cli):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_GET_BALANCE.format(cli)
    return subprocess.check_output(command).decode(DEFAULT_DECODE)

def listUnspent(cli):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_LIST_UNSPENT.format(cli)
    return subprocess.check_output(command).decode(DEFAULT_DECODE)
    
def getMasternodeOutputs(cli):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_MASTERNODE_OUTPUTS.format(cli)
    return subprocess.check_output(command).decode(DEFAULT_DECODE)
    
def generateMasternodeKey(cli):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_MASTERNODE_GENKEY.format(cli)
    return subprocess.check_output(command).decode(DEFAULT_DECODE).strip()

def masternodeStartAlias(cli, alias):
    """Wrapper function for the relevant RPC function call.
    
    Args:
        cli (str): Full path to cli binary associated with coin.
        alias (str): Alias associated with masternode to be started.
        
    Returns:
        String: String containing the command output.        
    """
    command = CLI_MASTERNODE_START_ALIAS.format(cli, alias)
    return subprocess.check_output(command).decode(DEFAULT_DECODE).strip()



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2018 Cosmos Coin Developers, https://cosmoscoin.co/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import time

import json
import paramiko

from io import BytesIO
from urllib.request import urlopen

DEFAULT_DECODE = "utf-8"

CHECK_RELEASE_COMMAND = "lsb_release -a"
UPDATE_TOOLS_COMMAND = "apt-get -y update && apt-get -y upgrade && apt-get -y install wget"

INSTALL_COIN_COMMAND = "mkdir -p /home/{0} && " \
                        "wget -qO- {1} | tar xvz --strip-components=1 -C /home/{0} && " \
                        "cp /home/{0}/bin/* /usr/local/bin"

GET_LATEST_GIT_RELEASE_COMMAND = "https://api.github.com/repos/{0}/{1}/releases/latest"
                        
IS_COIN_INSTALLED_COMMAND = "command -v {0}"
CHECK_IF_PROCESS_RUNNING_COMMAND_FMT = "ps cax | grep {0} > /dev/null"

def openChannel(server, username, password):
    """Open an SSH channel with the specified server.
    
    Args:
        server (str): The IP address of the server to connect to.
        username (str): The username to be used in the connection.
        password (str): The password associated with the user.
    
    Returns:
        Obj: A client object containing the state of the connection.
    """
    channel = paramiko.SSHClient()
    channel.set_missing_host_key_policy(paramiko.AutoAddPolicy)
    channel.connect(server, username=username, password=password)
    
    return channel
    
def closeChannel(channel):
    """Close a previously open SSH connection.
    
    Args:
        channel (obj): The client object returned by the open function.
    """
    channel.close()
    
def sendCommand(channel, command):
    """Send a comment across the channel and return the results.
    
    Args:
        channel (obj): The client object returned by the open function.
        command (str): The command to be executed.
    Returns:
        Str: A string object containing the command results.
    """        
    output = BytesIO()
    stdin, stdout, stderr = channel.exec_command(command)
    
    # Parse the partial command output while the command is running
    while not stdout.channel.exit_status_ready():
        while stdout.channel.recv_ready():
            output.write(stdout.channel.recv(1024))
        
        time.sleep(2)
    
    # Parse any remaining channel output
    while stdout.channel.recv_ready():
        output.write(stdout.channel.recv(1024))
        
    # Parse any error channel output
    while stderr.channel.recv_ready():
        output.write(stderr.channel.recv(1024))

    exitStatus = stdout.channel.recv_exit_status()
    
    if exitStatus != 0:
        raise ValueError("Command: \"{0}\" failed with status: {1}".format(command, exitStatus))
        
    return output.getvalue().decode(DEFAULT_DECODE)

def createFileWithContents(server, username, password, filePath, data):
    """Create a file on the specified server with the contents provided.
    
    Args:
        server (str): The IP address of the server to connect to.
        username (str): The username to be used in the connection.
        password (str): The password associated with the user.
        filePath (str): The relative path of the file to be written (relative to $HOME)
        data (str): The data to be written into the file.
    """        
    directory = os.path.dirname(filePath)
    
    # Open SSH connection
    channel = openChannel(server, username, password)
    sftp = channel.open_sftp()

    # Create directory if necessary
    try:
        sftp.mkdir(directory)
    except IOError:
        pass
      
    # Transfer and replace the file contents
    try:
        with sftp.open(filePath, 'w') as f:
            f.write(data)
    finally:
        # Close ssh connection
        closeChannel(channel)
        
def checkRelease(channel, codeName):
    """Check and verify the release on the VPS server.

    Args:
        channel (obj): The client object returned by the open function.
        codeName (str): The expected release code name to verify.
    """    
    print("Checking VPS operating system release..\n")
    output = sendCommand(channel, CHECK_RELEASE_COMMAND)
    
    print(output)
    if codeName not in output:
        raise ValueError("Unsupported operating system, refer to documentation for more info")

def checkDaemonNotRunning(channel, daemonName):
    """Check that the daemon is not currently running. This is done to avoid potential problems in the future.

    Args:
        channel (obj): The client object returned by the open function.
        daemonName (str): The name of the daemon to check.
    """
    print("Checking that daemon is not currently running on VPS..")
    if isProcessRunning(channel, daemonName):
        message = "The \'{0}\' daemon is currently running, please stop it and try again".format(daemonName)
        raise ValueError(message)
        
    print("")
        
def updateTools(channel):
    """Update the installation and tools on the VPS.
    
    Note: The output won't be printed until the update is complete.

    Args:
        channel (obj): The client object returned by the open function.
    """    
    print("Updating tools on VPS..\n")
    print(sendCommand(channel, UPDATE_TOOLS_COMMAND))

def getInstallCommand(coinName, gitOwner, gitProject, namePattern):
    """Determine the installation command based on the coin's latest release.
    
    Args:
        coinName (str): Name of the coin to install.
        gitOwner (str): The Git owner of the project.
        gitProject (str): The name of the project in Git.
        namePattern (str): A pattern to be used to identify the release version to get.
        
    Returns:
        String: A string containing the install command to be executed.
    """
    print("Get the latest masternode release")
    
    url = GET_LATEST_GIT_RELEASE_COMMAND.format(gitOwner, gitProject)
    releases = json.load(urlopen(url))

    matches = [release for release in releases["assets"] if namePattern in release["name"]]
        
    if len(matches) != 1:
        message = "Unexpected number of matches: {0} for the specified pattern {1}, please check configuration".format(len(txValues), namePattern)
        raise ValueError(message)
        
    downloadUrl = matches[0]["browser_download_url"]    
    print("Latest release found: {0}\n".format(downloadUrl))
    return INSTALL_COIN_COMMAND.format(coinName, downloadUrl)
    
def installMasternode(coinName, channel, daemonName, installCommand):
    """Install the masternode binaries on the VPS.

    Args:
        channel (obj): The client object returned by the open function.
        daemonName (str): The name of the daemon binary associated with the installation.
        installCommand (str): The full command to be executed for installation.
    """
    print("Installing masternode on VPS..")
    print("Install command:\n\n{0}\n".format(installCommand))
    
    try:
        # This command will fail if masternode binaries already installed
        sendCommand(channel, IS_COIN_INSTALLED_COMMAND.format(daemonName))
    except:
        print(sendCommand(channel, installCommand))
    else:
        print("{0} is already installed.. skipping installation.\n".format(coinName))    

def isProcessRunning(channel, processName):
    """Check if the specified process is currently running.
    
    Args:
        channel (obj): The client object returned by the open function.
        processName (str): The name of the process to check.
    
    Returns:
        Boolean: True if the process is running, False otherwise.
    """
    try: 
        # If process is not running, an exception is raised
        command = CHECK_IF_PROCESS_RUNNING_COMMAND_FMT.format(processName)
        sendCommand(channel, command)
    except:
        return False
    else:
        return True

def createUser(channel, coinName):
    """Creates a user for the coin if necessary.
    
    Args:
        channel (obj): The client object returned by the open function.
        coinName (str): Name of the coin.
    """
    print("Creating user for masternode: {0}, if necessary..".format(coinName))
    
    try:
        # If this command fails, the user doesn't exist
        sendCommand(channel, "id -u {0}".format(coinName))
    except:
        # Create user
        sendCommand(channel, "useradd {0}".format(coinName))
 
    print("")
 
def sendSingleCommand(server, user, password, command):
    """Wrapper function to open a connection and execute a single command.
    
    Args:
        server (str): The IP address of the server to connect to.
        username (str): The username to be used in the connection.
        password (str): The password associated with the user.
        command (str): The command to be executed.
        
    Returns:
        String: String containing the command output.
    """
    # Open SSH connection
    channel = openChannel(server, user, password)
    
    output = ""
    try:
        output = sendCommand(channel, command)
    finally:
        # Close ssh connection
        closeChannel(channel)

    return output
        
def setup(server, user, password, config):
    """Program entry point. This function will setup the VPS per the coin requirements.

    Args:
        server (str): The IP address of the server to connect to.
        username (str): The username to be used in the connection.
        password (str): The password associated with the user.
        config (dict): Dictionary containing the options parsed by the utility.
    """
    # Open SSH connection
    channel = openChannel(server, user, password)
    
    try:
        # Check that OS is the right version
        checkRelease(channel, config["VPS"]["UbuntuCodename"])
        
        # Check that daemon is not running
        checkDaemonNotRunning(channel, config["Coin"]["Daemon"])
    
        # Update OS tools
        updateTools(channel)
    
        # Install masternode
        installCommand = getInstallCommand(config["Coin"]["Name"], config["Git"]["Owner"], config["Git"]["Project"], config["Git"]["NamePattern"])
        installMasternode(config["Coin"]["Name"], channel, config["Coin"]["Daemon"], installCommand)
        
        # Create user for masternode
        createUser(channel, config["Coin"]["Name"])
    finally:
        # Close ssh connection
        closeChannel(channel)



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2018 Cosmos Coin Developers, https://cosmoscoin.co/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from setuptools import setup

setup(name="Cosmos Coin Masternode Setup",
      version="1.1.0",
      description="End to end script to setup a masternode",
      url="https://github.com/mtmonte/MasternodeSetup",
      author="mtmonte",
      author_email="cosmos.mtmonte@gmail.com",
      keywords="bitcoin cosmos coin crypto cryptocurrency masternode",
      license="MIT",
      packages=["MasternodeSetup"],
      python_requires=">=3.3.0",
      install_requires=["paramiko"],
      entry_points={"console_scripts": ["cosmos-masternode-setup=MasternodeSetup.command_line:main"]},
      include_package_data=True,
      zip_safe=False)      